package collections

import (
	"fmt"

	"github.com/cloudfoundry-incubator/notifications/v2/models"
)

type Sender struct {
	ID       string
	Name     string
	ClientID string
}

type sendersRepository interface {
	Insert(conn models.ConnectionInterface, sender models.Sender) (insertedSender models.Sender, err error)
	List(conn models.ConnectionInterface, clientID string) (retrievedSenderList []models.Sender, er error)
	Get(conn models.ConnectionInterface, senderID string) (retrievedSender models.Sender, err error)
	GetByClientIDAndName(conn models.ConnectionInterface, clientID, name string) (retrievedSender models.Sender, err error)
}

type SendersCollection struct {
	repo sendersRepository
}

func NewSendersCollection(repo sendersRepository) SendersCollection {
	return SendersCollection{
		repo: repo,
	}
}

func (sc SendersCollection) Set(conn ConnectionInterface, sender Sender) (Sender, error) {
	model, err := sc.repo.Insert(conn, models.Sender{
		Name:     sender.Name,
		ClientID: sender.ClientID,
	})
	if err != nil {
		switch err.(type) {
		case models.DuplicateRecordError:
			model, err = sc.repo.GetByClientIDAndName(conn, sender.ClientID, sender.Name)
			if err != nil {
				return Sender{}, PersistenceError{err}
			}
		default:
			return Sender{}, PersistenceError{err}
		}
	}

	return Sender{
		ID:       model.ID,
		Name:     model.Name,
		ClientID: model.ClientID,
	}, nil
}

func (sc SendersCollection) List(conn ConnectionInterface, clientID string) ([]Sender, error) {
	senderList := []Sender{}

	models, err := sc.repo.List(conn, clientID)
	if err != nil {
		return senderList, PersistenceError{err}
	}

	for _, model := range models {
		sender := Sender{
			ID:       model.ID,
			Name:     model.Name,
			ClientID: model.ClientID,
		}

		senderList = append(senderList, sender)
	}

	return senderList, nil
}

func (sc SendersCollection) Get(conn ConnectionInterface, senderID, clientID string) (Sender, error) {
	model, err := sc.repo.Get(conn, senderID)
	if err != nil {
		switch e := err.(type) {
		case models.RecordNotFoundError:
			return Sender{}, NotFoundError{e}
		default:
			return Sender{}, PersistenceError{err}
		}
	}

	if clientID != model.ClientID {
		return Sender{}, NotFoundError{fmt.Errorf("Sender with id %q could not be found", senderID)}
	}

	return Sender{
		ID:       model.ID,
		Name:     model.Name,
		ClientID: model.ClientID,
	}, nil
}
